#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import sys
import json
import requests
import traceback
import shelve

from langchain.tools.base import BaseTool

def get_area_code(area_name) -> None:
    if not area_name:
        return None

    api_key = os.environ.get('WEATHER_API_KEY', None)
    if not api_key:
        return None

    url = 'https://eolink.o.apispace.com/456456/function/v001/city?items=10&area=china&withPoi=false&location=' + area_name
    try:
        resp = requests.get(url, headers={
            'X-APISpace-Token': api_key,
            'Authorization-Type':'apikey',
        })
        if resp.status_code >= 300:
            print(f"Request weather api failed, http status code:{resp.status_code}", file=sys.stderr)
            return None

        res = json.loads(resp.content)
        if type(res) is not dict or res.get('status', -1) != 0:
            print(f'Failed to request area codeof [{area_name}], response:{resp.content}', file=sys.stderr)
            return ''

        area_code = res['areaList'][0]['areacode'].strip()
        #print(f'Got area code [{area_code}] of [{area_name}] ...', file=sys.stderr)
        return area_code
    except Exception as e:
        traceback.print_exc(file=sys.stderr)
        return None

def get_daily_weather_forecast(area_name):
    area_name = area_name.strip(' "')
    if not area_name:
        return []

    api_key = os.environ.get('WEATHER_API_KEY', None)
    if not api_key:
        return []

    #print(f"[get_daily_weather_info] area_name:{area_name}", file=sys.stderr)
    area_code = get_area_code(area_name)
    if not area_code:
        return []

    url = 'https://eolink.o.apispace.com/456456/weather/v001/day?days=5&areacode=' + area_code
    try:
        resp = requests.get(url, headers={
            'X-APISpace-Token': api_key,
            'Authorization-Type':'apikey',
        })
        if resp.status_code >= 300:
            print(f"Request weather api failed, http status code:{resp.status_code}", file=sys.stderr)
            return []
        res = json.loads(resp.content)
        if type(res) is not dict or len(res.get('result', [])) == 0:
            return []
        res_list = []
        day_names = ['明天', '后天', '大后天', '4天后']
        fcsts = res['result']['daily_fcsts']
        for i in range(len(fcsts)):
            item = fcsts[i]
            res_list.append({
                'date':item['date'],
                'date_name': day_names[i] if i < len(day_names) else f'{i+1}天后',
                'text_day':item['text_day'],
                'text_night':item['text_night'],
                'high_temp':item['high'],
                'low_temp':item['low'],
                'wind_class_day':item['wc_day'],
                'wind_class_night':item['wc_night'],
                'wind_direction_day':item['wd_day'],
                'wind_direction_night':item['wd_night'],
            })
        return res_list
    except Exception as e:
        traceback.print_exc(file=sys.stderr)
        return None

if __name__ == '__main__':
    query = sys.argv[1]
    res = get_daily_weather_forecast(query)
    print(json.dumps(res, indent=4, ensure_ascii=False))




import os
import sys
import json
import traceback

from langchain.agents import tools
from langchain.llms.base import BaseLLM

from reasoningchain.executor.sandbox import execute_python
from reasoningchain.cache.disk_cache import disk_cache
from reasoningchain import api

CACHE_DIR = os.environ.get('REASONING_CHAIN_CACHE_PATH', os.path.join(os.environ['HOME'], '.cache'))
if not os.path.exists(CACHE_DIR):
    os.mkdir(CACHE_DIR)

CACHE_DIR = os.path.join(CACHE_DIR, 'reasoning_chain')
if not os.path.exists(CACHE_DIR):
    os.mkdir(CACHE_DIR)

custom_tool_map = {}
def custom_tool(name:str, description:str):
    def decorator(func:callable):
        custom_tool_map[name] = {
            'name': name,
            'description': description,
            'func': func,
        }
        return func
    return decorator

def get_all_custom_tool_names():
    return [k for k in custom_tool_map]

def get_all_custom_tool_info():
    return {name:info["description"] for name, info in custom_tool_map.items()}

def load_tool(name:str, callback=None, mock_output=None):
    tool_info = custom_tool_map[name]
    tool_func = custom_tool_map[name]['func']
    def func(query:str):
        if mock_output:
            if callback:
                return callback(query, mock_output)
            return mock_output
        if callback is not None:
            return tool_func(query, callback)
        else:
            return tool_func(query)
    return tools.Tool(
        name = name,
        description = tool_info['description'],
        func = func,
    )

def load_custom_tools(tool_names:list, callbacks:dict={}):
    tools = []
    for name in tool_names:
        callback = callbacks.get(name, None) if callbacks else None
        tool = load_tool(name, callback)
        tools.append(tool)
    return tools

def get_all_tool_names():
    from langchain import agents
    return get_all_custom_tool_names() + agents.get_all_tool_names()

def load_tools(tool_names:list, tool_callbacks:dict=None, llm:BaseLLM=None) -> list:
    from langchain import agents

    custom_tool_names = []
    predefined_tool_names = []
    all_agent_tool_names = agents.get_all_tool_names()
    for name in tool_names:
        name = name.strip()
        if not name:
            continue
        if name in get_all_custom_tool_names():
            custom_tool_names.append(name)
        elif name in all_agent_tool_names:
            predefined_tool_names.append(name)
        else:
            raise RuntimeError(f"Unknown tool name:[{name}]")

    tools = []
    if len(custom_tool_names) > 0:
        tools += load_custom_tools(custom_tool_names, tool_callbacks)

    if len(predefined_tool_names) > 0:
        from langchain.llms import OpenAI, OpenAIChat
        if llm is None:
            llm = OpenAIChat(temperature=0)
        tools += agents.load_tools(predefined_tool_names, llm)

    return tools





import os
import sys
import click

@click.command()
@click.option("--port", "-p", "port", type=int, default=8501, required=False)
def main(port:int) -> int:
    import signal
    import subprocess

    def signal_handler(signum, frame):
        print(f"Received signal {signum}. Terminating child process...")
        os.kill(os.getpid(), signal.SIGTERM)

    signal.signal(signal.SIGINT, signal_handler)

    uipath = os.path.join(os.path.abspath(__file__ + '/..'), "stui.py")

    try:
        p = subprocess.Popen(["streamlit", "run", uipath, "--server.port", str(port)])
        print("Web UI process:", p, file=sys.stderr)
        while p.poll() is None:
            p.wait()
        print(f"WebUI process exited with code:{p.returncode}", file=sys.stderr)
        return p.returncode
    except KeyboardInterrupt:
        print("\nMain process interrupted. Terminating...", file=sys.stderr)
        os.kill(p.pid, signal.SIGTERM) 
        return -1

if __name__ == '__main__':
    main()




__version__ = '0.0.22'

import os
import sys

from typing import Union

from . import custom_tools
from . import index
from . import cache
from . import api
from ._run import run

from .custom_tools.base import get_all_tool_names, load_tools




#!/usr/bin/env python3

import os
import sys
import subprocess

def execute_python(codes:str) -> str:
    codes = codes.strip()
    if codes.startswith('```'):
        parts = codes.split('\n', 1)
        if len(parts) == 1:
            codes = codes.strip('`')
        else:
            codes = parts[1].rstrip('`')
    codes = codes.strip().strip('`')
    print("execute_python:{\n%s\n}" % codes)

    cp = subprocess.Popen(["python3", "-c", codes], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    out = cp.communicate()[0].decode().strip()
    if cp.returncode != 0:
        out += f'\n\nProcess exited abnormally with exit code {cp.returncode}'
    return out

if __name__ == '__main__':
    codes = ''
    for line in sys.stdin:
        codes += line
    execute(codes)




#!/usr/bin/env python3

import os
import setuptools
from setuptools import find_packages, setup

pkg_dir = os.path.abspath(__file__ + '/..')

setup(
    name='reasoningchain',
    version='0.0.22',
    author='duer',
    long_description=open(os.path.join(pkg_dir, "README.md")).read(),
    long_description_content_type='text/markdown',
    python_requires = '>=3.8.0',
    packages = find_packages(),
    entry_points={
        "console_scripts": [
            "reasoningchainui=reasoningchain.webui.webui:main",
            "reasoningchain=reasoningchain._run:cli"
        ],
    },
    install_requires=[
        "langchain",
        "click",
        "openai",
        "LAC",
        "tqdm",
        "faiss-cpu",
        "streamlit",
    ],
)




#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import sys
import json
import time
import regex
import logging
import traceback
import pandas as pd
import numpy as np
from io import BytesIO
from langchain import agents
from langchain.llms import OpenAI,OpenAIChat
from langchain.chains import LLMChain
from langchain.prompts import PromptTemplate
from langchain.chains import SimpleSequentialChain
from langchain.tools.base import BaseTool
from langchain.agents import load_tools
from langchain.agents import initialize_agent
from langchain.agents import mrkl
from langchain.callbacks import set_handler as lc_set_handler
from langchain.callbacks.base import BaseCallbackHandler
#from langchain.callbacks.streamlit import StreamlitCallbackHandler

from reasoningchain.custom_tools import custom_tool, get_all_custom_tool_info, get_all_custom_tool_names, get_all_tool_names
from reasoningchain._run import run as run_rc

def render():
    import streamlit as st
    st.set_page_config(
        page_title="Reasonging Chain Is All You Need",
        page_icon="🧠",
        layout="wide",
        initial_sidebar_state="expanded",
    )

    def multiline_markdown(text):
        for line in text.split('\n'):
            line = line.rstrip()
            st.markdown(line)

    class CustomCallbackHandler(BaseCallbackHandler):
        @property
        def always_verbose(self) -> bool:
            """Whether to call verbose callbacks even if verbose is False."""
            return True

        def on_llm_start(self, serialized, prompts, **kwargs):
            print('on_llm_start():', serialized, prompts, kwargs)
            with st.expander(f"Calling LLM: {serialized['name']}"):
                for prompt in prompts:
                    st.write(prompt)

        def on_llm_new_token(self, token, **kwargs):
            print('on_llm_new_token():', token, kwargs)

        def on_llm_end(self, response, **kwargs):
            pass
            #print('on_llm_end():', response, kwargs)

        def on_llm_error(self, error, **kwargs):
            print('on_llm_error():', error, kwargs)

        def on_chain_start(self, info, inputs, **kwargs):
            return
            try:
                input_str = inputs["input"].strip().replace('\n', ' ')
                print(f'\x1b[33m{info["name"]}\x1b[0m(\x1b[32m{input_str}\x1b[0m)')
                st.markdown(f'{info["name"]}(`{input_str}`)')
            except:
                pass

        def on_chain_end(self, outputs, **kwargs):
            with st.container():
                if 'answer' in outputs:
                    text = outputs['answer']
                elif 'text' in outputs:
                    text = outputs['text']
                elif 'output' in outputs:
                    text = outputs['output']
                else:
                    text = ''
                st.markdown(f'Chain output:`{text}`')

        def on_chain_error(self, error, **kwargs):
            error = str(error)
            print(f'Chain error:\x1b[31m{error}\x1b[0m')
            st.caption(f'Chain error::red[{error}]')

        def on_tool_start(self, info, input_str: str, **kwargs):
            pass
            #raise RuntimeError("on_tool_start()")
            #for c in '" \n':
            #    input_str = input_str.strip(c)
            #print(f'Action:\x1b[33m{info["name"]}\x1b[0m(\x1b[32m{input_str}\x1b[0m)')
            #st.markdown(f'Action: **{info["name"]}**(`{input_str}`)')

        def on_tool_end(self, output: str, **kwargs):
            output = output.strip(' "')
            print(f'Output:\x1b[32m{output}\x1b[0m')
            if ' url:http' in output:
                pass
            elif ' gps:{' in output:
                try:
                    gpsv = output.split(':',1)[1]
                    gps = json.loads(gpsv)
                    st.markdown(f'gps:`{gpsv}`')
                    df = pd.DataFrame(
                        np.array([[gps['latitude'], gps['longitude']]]),
                        columns=['lat', 'lon'])
                    st.map(df)
                except Exception as e:
                    st.error(e)
                    traceback.print_exc(file=sys.stderr)
            else:
                st.markdown(f'**Output:** `{output}`')

        def on_tool_error(self, error, **kwargs):
            print(f'Tool error:\x1b[31m{error}\x1b[0m')

        def on_text(self, text: str, **kwargs):
            pass
            #print('Text:', text)

        def on_agent_action(self, action, **kwargs):
            try:
                tool_input = action.tool_input.strip().replace('\n', ' ')
                print(f'Action:\x1b[33m{action.tool}\x1b[0m(\x1b[32m{tool_input}\x1b[0m)')
                show_text = tool_input if f'`{tool_input}`' else ''
                st.markdown(f'Action: **{action.tool}** ({show_text})')
            except:
                print(action, kwargs)

        def on_agent_finish(self, finish, **kwargs):
            pass
            #print('on_agent_finish():', finish, kwargs)

    sthandler = CustomCallbackHandler()
    lc_set_handler(sthandler)

    @custom_tool(
        name = "TableTool",
        description = (
            "A tool for displaying data as tables. "
            "Useful for when you need to display some data in table format. "
            "Input should be the data in json format. "
        )
    )
    def table_tool(input_data:str, callback:callable=None) -> str:
        try:
            chart_data = json.loads(input_data)
            st.table(
                pd.DataFrame(
                    np.array([chart_data['x_data'], chart_data['y_data']]).T,
                    columns=[chart_data['x_name'], chart_data['y_name']],
                ))
            return "A table containing the data has been drawed successfully as you wish. "
        except Exception as e:
            traceback.print_exc(file=sys.stderr)
            st.error(f'Line chart error:{e}')
            return 'Some error occurred when drawing the line chart. '

    @custom_tool(
        name = "DrawLineChart",
        description = (
            "A tool for drawing line charts. "
            "Useful for when you need to draw a line chart with some data. "
            "Input should be the data in json format. "
        )
    )
    def draw_line_chart_tool(input_data:str, callback:callable=None) -> str:
        try:
            st.write('Drawing line chart')
            chart_data = json.loads(input_data)
            st.line_chart(
                data=pd.DataFrame(
                    np.array([chart_data['x_data'], chart_data['y_data']]).T,
                    columns=[chart_data['x_name'], chart_data['y_name']],
                ),
                x=chart_data['x_name'],
                y=chart_data['y_name'])
            return "A line chart has been drawed successfully. "
        except Exception as e:
            traceback.print_exc(file=sys.stderr)
            st.error(f'Line chart error:{e}')
            return 'Some error occurred when drawing the line chart. '

    def on_weather_result(query, weather_forcasts):
        if not weather_forcasts:
            return weather_forcasts
        #st.json(weather_forcasts)
        cols = st.columns(len(weather_forcasts))
        x, y_high, y_low = [], [], []
        for i in range(len(cols)):
            weather = weather_forcasts[i]
            x.append(weather['date'])
            y_high.append(float(weather['high_temp']))
            y_low.append(float(weather['low_temp']))
            cols[i].metric(weather['date'], weather['text_day'])

        #st.line_chart(
        #    data=pd.DataFrame(
        #        np.array([x, y_high, y_low]).T,
        #        columns=['日期', '最高气温', '最低气温'],
        #    ),
        #    x='日期',
        #    y=['最高气温','最低气温'])
        return weather_forcasts

    def on_image_result(query, url):
        st.image(url)
        return url

    def on_play_music(query, music_info):
        if not music_info:
            return music_info
        if music_info['type'] == 'video':
            st.video(music_info['url'])
        elif music_info['type'] == 'audio':
            st.audio(music_info['url'])
        return music_info

    def on_search_result(query, result):
        return result

    def on_python_executor(input_text, result):
        codes = input_text.strip()
        if codes.startswith('```'):
            codes = codes.split('\n')[1].strip()
        codes = '```Python\n' + codes
        if not codes.endswith('```'):
            codes += '\n```'

        st.caption("Python codes:")
        st.markdown(codes)
        st.markdown(f'Execution output: `{result}`')

    tool_callbacks = {
        "BaiduSearchText": on_search_result,
        "OpenaiAIPainter": on_image_result,
        'SearchImage': on_image_result,
        "WeatherForcast": on_weather_result,
        "PythonExecutor": on_python_executor,
    }

    col_left, col_main, col_rght = st.columns([25,50,25])

    with col_left:
        st.empty()

    with col_main:
        st.title(":blue[ReasoningChain] Is All You Need")
        tab_chat, tab_toolbox, tab_help = st.tabs(["问答", "工具箱", "帮助"])

        #img_file_buffer = st.camera_input("Take a picture")
        with tab_chat:
            agent_name = st.selectbox("Agent:", ["zero-shot-react-description"])
            tool_options = st.multiselect(
                    "选择工具（tools）",
                    get_all_tool_names(),
                    ["BaiduSearchText","GoogleSearchImage","GoogleSearchMap","OpenaiAIPainter","WeatherForecast"])

            def assistant(query):
                query = query.strip()
                if query:
                    text = run_rc(query, tool_names=tool_options)
                    st.markdown("#### Final Answer:")
                    re = regex.compile('.* image.*: ?(https?://[^ ]*)')
                    for ans in text.split('\n'):
                        urls = re.findall(ans)
                        if len(urls) > 0:
                            for url in urls:
                                pos = ans.find(url)
                                if pos < 0:
                                    continue
                                part1 = ans[:pos]
                                part2 = ans[pos+len(url):]
                                #tts1 = get_tts(part1)
                                #tts2 = get_tts(part2)

                                st.markdown(f'{part1}')
                                #st.audio(BytesIO(tts1))
                                st.image(url)
                                st.markdown(f'{part2}')
                                #st.audio(tts2)
                        else:
                            st.markdown(f"{ans}")
                            #tts = get_tts(ans)
                            #st.audio(tts)


            def on_submit(user_question):
                t = time.strftime("%Y-%m-%d %H:%M:%S")
                with st.spinner('努力思考中 ...'):
                    with st.container():
                        assistant(user_question)

            with st.form("el_input_block"):
                user_question = st.text_input(label="问题：", placeholder="在此输入问题")
                # Every form must have a submit button.
                submitted = st.form_submit_button("提交", use_container_width=True)
                if submitted:
                    on_submit(user_question)

            with st.expander("Edit Agent's LLM Prompt:"):
                st.markdown("#### Zero Shot React Agent")
                prompt_prefix = st.text_area(label='prefix:', value=mrkl.prompt.PREFIX)
                prompt_format_instructions = st.text_area(label='format_instructions:', value=mrkl.prompt.FORMAT_INSTRUCTIONS)
                prompt_suffix = st.text_area(label='suffix:', value=mrkl.prompt.SUFFIX)
                def on_save_prompt():
                    mrkl.prompt.PREFIX = prompt_prefix
                    mrkl.prompt.FORMAT_INSTRUCTIONS = prompt_format_instructions
                    mrkl.prompt.SUFFIX = prompt_suffix
                    st.write('保存成功!')
                st.button(label="保存", on_click=on_save_prompt, use_container_width=True)

        with tab_toolbox:
            st.caption("predefined tools: https://langchain.readthedocs.io/en/latest/modules/agents/tools.html")
            with st.container():
                st.header("Custom Tools")
                for name, description in get_all_custom_tool_info().items():
                    with st.container():
                        st.markdown(f"#### {name}")
                        st.markdown(f'{description}')
        with tab_help:
            st.write("Reasoning Chain is all you need, try it!")

    with col_rght:
        st.empty()

if __name__ == '__main__':
    render()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import sys
import json
import requests
import traceback
import asyncio
from typing import Union

from reasoningchain.cache.disk_cache import disk_cache

def generate_image(prompt:str):
    prompt = prompt.strip()
    if not prompt:
        return []

    req = {
        "prompt": prompt,
        "n": 1,
        "size": "512x512"
    }
    res = request_api('/images/generations', req)
    if not res:
        return []

    res = json.loads(response.content)
    urls = []
    for item in res['data']:
        urls.append(item['url'])
    return urls

def get_embedding(text:str):
    try:
        text = text.strip()
        if not text:
            return None

        params = {
            "model": "text-embedding-ada-002",
            "input": text
        }
        res = request_api('/embeddings', params)
        if not res.get('data', None):
            return None
        return res['data'][0]['embedding']
    except Exception as e:
        traceback.print_exc(file=sys.stderr)
        return None 

def chat(messages:Union[list,str], max_tokens=500, temperature=0.5):
    if isinstance(messages, str):
        messages = [{
            'role':'user',
            'content':messages,
        }]

    req = {
         "model": "gpt-3.5-turbo",
         "temperature": temperature,
         "max_tokens": max_tokens,
         "messages": messages
    }
    res = request_api("/chat/completions", req)
    if not res:
        return res
    print('closeai token usage:', json.dumps(res['usage'], ensure_ascii=False), file=sys.stderr)
    answer = res['choices'][0]['message']['content']
    return answer

@disk_cache(cache_path=os.path.join(os.environ['HOME'], '.cache/closeai/chat'), expire_time=864000)
def chat_with_cache(messages:Union[list,str], max_tokens=500, temperature=0.5):
    return chat(messages, max_tokens, temperature)

@disk_cache(cache_path=os.path.join(os.environ['HOME'], '.cache/closeai/text_embedding'), expire_time=8640000)
def get_embedding_with_cache(text:str):
    return get_embedding(text)

async def aget_embedding_with_cache(text:str):
    return get_embedding_with_cache(text)

def batch_get_embeddings(text_list:list, batch_size=8) -> list:
    res = []
    loop = asyncio.get_event_loop()
    try:
        for i in range(0, len(text_list), batch_size):
            tasks = []
            for text in text_list[i:i+batch_size]:
                tasks.append(asyncio.ensure_future(aget_embedding_with_cache(text)))

            results = loop.run_until_complete(asyncio.wait(tasks))
            for item in results[0]:
                res.append(item.result())
    except Exception as e:
        print(e, file=sys.stderr)
    loop.close()
    return res

def request_api(path, params):
    url = os.getenv("OPENAI_API_BASE")
    if not url:
        url = "https://api.openai.com/v1"
    url += path
    token = os.getenv('OPENAI_API_KEY')
    response = requests.post(url, json=params, headers={
        'Content-Type': 'application/json',
        'Authorization': f'Bearer {token}',
    })
    if response.status_code >= 300:
        print(f'Request "{path}" failed with status:{response.status_code}', file=sys.stderr)
        return None

    return json.loads(response.content)

if __name__ == '__main__':
    cmd = sys.argv[1]
    if cmd in ['emb', 'embedding']:
        text_list = []
        for line in sys.stdin:
            text = line.strip()
            if text:
                text_list.append(text)

        res = batch_get_embeddings(text_list)
        print(res)
        print(len(res))




#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""Wrapper around OpenAI APIs."""
from typing import Any, Dict, List, Mapping, Optional
import requests
import json
import os
import sys

from typing import Union, Tuple

from pydantic import BaseModel, Extra, Field, root_validator

from langchain.llms.base import BaseLLM
from langchain.schema import Generation, LLMResult
from langchain.utils import get_from_dict_or_env

class DuLLM(BaseLLM, BaseModel):
    client: Any  #: :meta private:
    model_name: str = "glm-10b"
    """Model name to use."""
    temperature: float = 0.7
    """What sampling temperature to use."""
    max_tokens: int = 256
    """The maximum number of tokens to generate in the completion.
    -1 returns as many tokens as possible given the prompt and
    the models maximal context size."""
    top_p: float = 1
    """Total probability mass of tokens to consider at each step."""
    frequency_penalty: float = 0
    """Penalizes repeated tokens according to frequency."""
    presence_penalty: float = 0
    """Penalizes repeated tokens."""
    n: int = 1
    """How many completions to generate for each prompt."""
    best_of: int = 1
    """Generates best_of completions server-side and returns the "best"."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""
    openai_api_key: Optional[str] = None
    batch_size: int = 20
    """Batch size to use when passing multiple documents to generate."""
    request_timeout: Optional[Union[float, Tuple[float, float]]] = None
    """Timeout for requests to OpenAI completion API. Default is 600 seconds."""
    logit_bias: Optional[Dict[str, float]] = Field(default_factory=dict)
    """Adjust the probability of specific tokens being generated."""
    max_retries: int = 6
    """Maximum number of retries to make when generating."""
    streaming: bool = False
    """Whether to stream the results or not."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.ignore

    @root_validator(pre=True)
    def build_extra(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Build extra kwargs from additional params that were passed in."""
        all_required_field_names = {field.alias for field in cls.__fields__.values()}

        extra = values.get("model_kwargs", {})
        for field_name in list(values):
            if field_name not in all_required_field_names:
                if field_name in extra:
                    raise ValueError(f"Found {field_name} supplied twice.")
                extra[field_name] = values.pop(field_name)
        values["model_kwargs"] = extra
        return values

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        values["client"] = '-'
        if values["streaming"] and values["n"] > 1:
            raise ValueError("Cannot stream results when n > 1.")
        if values["streaming"] and values["best_of"] > 1:
            raise ValueError("Cannot stream results when best_of > 1.")
        return values

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling OpenAI API."""
        normal_params = {
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            "top_p": self.top_p,
            "n": self.n,
            "best_of": self.best_of,
            "request_timeout": self.request_timeout,
            "logit_bias": self.logit_bias,
        }
        return {**normal_params, **self.model_kwargs}

    def _generate(
        self, prompts: List[str], stop: Optional[List[str]] = None
    ) -> LLMResult:
        choices = []
        for prompt in prompts:
            res = self.generate_answer(prompt, stop)
            choices.append(res)
        return self.create_llm_result(choices, prompts, {})

    async def _agenerate(
        self, prompts: List[str], stop: Optional[List[str]] = None
    ) -> LLMResult:
        """Call out to OpenAI's endpoint async with k unique prompts."""
        choices = []
        for prompt in prompts:
            res = self.generate_answer(prompt, stop)
            choices.append(res)
        return self.create_llm_result(choices, prompts, token_usage)

    def generate_answer(self, prompt:str, stop:list=[]) -> str:
        if not stop:
            stop = []
        req = {
            **self._default_params,
            "input":prompt,
            "model_type":self.model_name,
        }
        DU_LLM_API = os.environ['DU_LLM_API']

        resp = requests.post(DU_LLM_API, json=req)
        if resp.status_code >= 300:
            raise RuntimeError(f"Request DuLLM(self.model_name) failed with http status:{resp.status_code}")

        res = json.loads(resp.content)
        if res['status'] != 0:
            raise RuntimeError(f"Request DuLLM(self.model_name) failed with api status:{res['status']}, msg:{res['msg']}")
        
        answer = res['data']['output'].lstrip('<|startofpiece|>')
        finish_reason = "end"
        for item in stop:
            pos = answer.find(item)
            if pos >= 0:
                answer = answer[:pos+len(item)]
                finish_reason = "stop"
        return {"text":answer, "finish_reason":finish_reason}

    def create_llm_result(
        self, choices: Any, prompts: List[str], token_usage: Dict[str, int]
    ) -> LLMResult:
        """Create the LLMResult from the choices and prompts."""
        generations = []
        for i, prompt in enumerate(prompts):
            res = choices[i]
            print(res)
            generations.append(
                [
                    Generation(
                        text=res['text'],
                        generation_info=dict(
                            finish_reason=res['finish_reason'],
                            logprobs=1.0,
                        ),
                    )
                ]
            )
        llm_output = {"token_usage": token_usage, "model_name": self.model_name}
        return LLMResult(generations=generations, llm_output=llm_output)

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {**{"model_name": self.model_name}, **self._default_params}

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "dullm"

    def get_num_tokens(self, text: str) -> int:
        """Calculate num tokens with tiktoken package."""
        # calculate the number of tokens in the encoded text
        return len(text)

    def modelname_to_contextsize(self, modelname: str) -> int:
        if modelname == "glm-10b":
            return 4097
        elif modelname == "bloomz_7b1":
            return 4000
        else:
            return 4097

    def max_tokens_for_prompt(self, prompt: str) -> int:
        """Calculate the maximum number of tokens possible to generate for a prompt.

        Args:
            prompt: The prompt to pass into the model.

        Returns:
            The maximum number of tokens to generate for a prompt.

        Example:
            .. code-block:: python

                max_tokens = openai.max_token_for_prompt("Tell me a joke.")
        """
        num_tokens = self.get_num_tokens(prompt)

        # get max context size for model by name
        max_size = self.modelname_to_contextsize(self.model_name)
        return max_size - num_tokens

if __name__ == '__main__':
    prompt = "My name is "
    if len(sys.argv) > 1:
        prompt = sys.argv[1]

    llm = DuLLM(model_name="glm-10b", max_tokens=1000)

    answer = llm._generate([prompt])
    print(f"Answer: {answer}")



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import sys
import json
import random
import requests
import traceback
import sqlite3

from reasoningchain.executor.sandbox import execute_python
from reasoningchain.cache.disk_cache import disk_cache
from reasoningchain import api

from .base import custom_tool, CACHE_DIR

@disk_cache(cache_path=os.path.join(CACHE_DIR, "baidu_search_text.db"), expire_time=86400*2)
def baidu_search_text(query:str):
    return api.baidu.search_text(query)

@custom_tool(
    name = "BaiduSearchText",
    description = "A search engine. Useful for when you need to search textual information about facts and current events. Never use it to do math calculations. Input should be a query for the search engine."
)
def baidu_search_text_tool(query:str, callback:callable=None) -> str:
    query = query.strip()
    if not query:
        return "No invalid query. "
    result = baidu_search_text(query) # 空结果不缓存
    if result:
        if callback:
            result = callback(query, result)
        return '搜索结果：' + result.replace("\n", " ")
    return "Nothing is found. "

@disk_cache(cache_path=os.path.join(CACHE_DIR, "google_search_image.db"), expire_time=86400*7)
def google_search_image(query:str):
    return api.google.search_image(query)

@disk_cache(cache_path=os.path.join(CACHE_DIR, "google_search_map.db"), expire_time=86400*7)
def google_search_map(query:str, callback:callable=None):
    return api.google.search_map(query)

@custom_tool(
    name = "GoogleSearchImage",
    description = "A search engine for images. Useful for when you need to find an image or a picture from the web or you are asked what somebody or something looks like. Input should be a query for the image search engine."
)
def google_search_image_tool(query:str, callback:callable=None) -> str:
    query = query.strip()
    if not query:
        return "No invalid query. "
    results = google_search_image(query) # 空结果不缓存
    if not results:
        return "Nothing is found."
    url = results[int(random.random()*len(results))]["original"]
    if callback:
        url = callback(query, url)
    return "An image is found:" + url

@custom_tool(
    name = "GoogleSearchMap",
    description = "A search engine for locations on map. Useful for when you need to find the position of somewhere on the map. Input should be a query in Chinese for the map search engine."
)
def google_search_map_tool(query:str, callback:callable=None) -> str:
    query = query.strip()
    if not query:
        return "No invalid query. "
    loc = google_search_map(query) # 空结果不缓存
    if not loc:
        return "Nowhere is found."
    if callback is not None:
        loc = callback(query, loc)
    return 'I found the place, it\'s at %s, gps:{"latitude":%f, "longitude":%f}' % (loc['address'], loc['latitude'], loc['longitude'])

@custom_tool(
    name = "PythonExecutor",
    description = "A python code executor. Usefull for when you have a piece of python codes to execute. Input should be the codes to be executed. You can get the execution output by using print(). "
)
def python_executor_tool(codes:str, callback:callable=None):
    try:
        out = execute_python(codes)
        res = f'Execution output: {out}'
    except Exception as e:
        out = ''
        res = "Caught exception: " + str(e)
    if callback is not None:
        callback(codes, out)
    return res

@custom_tool(
    name = 'SearchStudentScores',
    description = "A search tool for query or calculating studentss examination scores. Useful for when you need to query or calcualate students examination scores. The scores data are stored in a MySQL database table named student_scores, and its columns includes (exam_year: year of examinations date. exam_date: examinations date.  teacher_name: teacher name of the examination course.  class_name: class name of the student.  student_name: student name. course_name: course name. score: score of the examination.) Input should be a SQL to retrieving or calculating the data you want. The output would be in json format. "
)
def search_student_scores_tool(sql:str, callback:callable=None) -> str:
    sql = sql.strip()
    if sql.endswith('"'):
        sql = sql[:-1]
    if not sql:
        return "I got empty input and can do nothing."

    try:
        conn = sqlite3.connect("/home/work/llmui/db/student_center.db")
        csr = conn.cursor()
        x = []
        y = []
        for date, score in csr.execute(sql):
            print(f'date:{date} score:{score}')
            x.append(date)
            y.append(score)
        res = {
            'x_name':'exam_date',
            'x_data': x,
            'y_name':'score',
            'y_data': y,
        }
        if callback is not None:
            res = callback(sql, res)
        return f'Here is the data you want in json format:' + json.dumps(res, ensure_ascii=False)
    except Exception as e:
        print("[SearchStudentScores::_run()] Caught exception:", e, file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
    return "I didn't find any data for your request."

class DBHelper:
    def __init__(self, dbname):
        self._dbname = dbname
        self._conn = sqlite3.connect(dbname)

    def execute(self, sql, commit=False):
        csr = self._conn.cursor()
        res = csr.execute(sql)
        if commit:
            self.commit()
        return res

    def commit(self):
        return self._conn.commit()

    def close(self):
        self._conn.close()


@disk_cache(cache_path=os.path.join(CACHE_DIR, "weather_forcast.db"), expire_time=7200)
def get_daily_weather_forecast(area_name):
    return api.weather.get_daily_weather_forecast(area_name)

@custom_tool(
    name = 'WeatherForecast',
    description = "A weather tool. Useful for when you need to answer questions about the weather or when you need to check what the weather will look like in the future. Input should be the date and the location in Chinese in format like this:(时间, 地点). "
)
def weather_forcast_tool(query:str, callback:callable=None):
    parts = query.split(',', 1)
    if len(parts) > 1:
        date, area_name = parts
    else:
        date = query
        area_name = ''
    area_name = area_name.strip(' "')
    date = date.strip(' "')
    if area_name.lower() == 'current location' or not area_name:
        area_name = '北京'
    print(f"[weather_forcast] area_name:{area_name}", file=sys.stderr)

    daily_weather_forcast = get_daily_weather_forecast(area_name)
    if not daily_weather_forcast:
        return 'Nothing is found. '

    if callback is not None:
        daily_weather_forcast = callback(query, daily_weather_forcast)

    weather = ''

    for i in range(len(daily_weather_forcast)):
        item = daily_weather_forcast[i]
        parts = [
            f'天气为{item["text_day"]}',
            f'最高气温{item["high_temp"]}摄氏度，最低气温{item["low_temp"]}摄氏度',
        ]
        if item.get('wc_day', None):
            parts.append(f'风速{item["wind_class_day"]}')
        if item.get('wd_day', None):
            parts.append(f'风向{item["wind_direction_day"]}')
        weather += item['date_name'] + '，'.join(parts) + '。'
    return weather

@custom_tool(
    name = "Scheduler",
    description = "A scheduler. Useful for when you need to query or manipulating schedules. Input should be the operation which should be one of [query, add, delete], time, content. "
)
def scheduler_tool(query:str, callback:callable=None) -> str:
    parts = query.split(',', 2)
    if len(parts) != 3:
        return "Sorry, I can not understand your input."

    op, tm, content = parts
    op = op.lower()
    if op == 'query':
        return 'Found a schedule on {tm.lower()}: travelling to Shanghai.'
    elif op == 'delete':
        return 'Schedule canceled.'
    elif op == 'add':
        return f'Schedule added on {tm.lower()}.'
    return "Sorry, operation \"{op}\" is not supported."


def _search_music(name):
    musics = {
        '爷爷的情书': {
            'type':'video',
            'alias':['A love letter from Grandpa', "Grandpa's Love Letter"],
            'url':'https://topcoder.top/static/videos/yydqs.mp4',
        },
        '半兽人': {
            'type':'video',
            'alias':['Half Beast', 'Half Orc'],
            'url':'https://topcoder.top/static/videos/banshouren.mp4',
        },
    }
    print(f"music name:[\x1b[32m{name}\x1b[0m]", file=sys.stderr)
    if name in musics:
        return musics[name]
    cname = name.lower().replace('-', ' ').strip()
    for k, v in musics.items():
        if name in k:
            return v
        for alias in v.get('alias', []):
            alias = alias.lower().replace('-', ' ').strip()
            if alias == cname:
                return v
    return None

@custom_tool(
    name = "MusicPlayer",
    description = (
        "A music player. "
        "Useful for when you have a piece of music or a song to search and play. "
        "Input should be name of the music or the song. "
    )
)
def music_player_tool(query:str, callback:callable=None) -> str:
    query = query.strip(' \n')
    if query.endswith('"'):
        query = query[:-1]
    name = query
    music = _search_music(name)
    if music:
        if callback:
            callback(query, music)
        return 'I found the music, and player starts playing it.'
    return "Sorry, I didn't found the music. Try again with a Chinese music name. "

@custom_tool(
    name = "BookTicket",
    description = (
        "A tool for booking tickets."
        "Useful for when you need to book or buy tickets. "
        "Input should be ticket type, destination, time. The ticket type should be one of [train, plane, entrance]. "
    )
)
def book_ticket_tool(query:str, callback:callable=None) -> str:
    parts = query.split(',', 2)
    if len(parts) != 3:
        return "Sorry, I can not understand your input."

    tt, destination, tm = parts
    tt = tt.lower()
    return 'The ticket has beend ordered successfully. '

@custom_tool(
    name = "HomeCircumstances",
    description = (
        "A tool for querying circumstances of home."
        "Useful for when you need to check the temperature of a room. "
        "Input should be the room name. "
    )
)
def home_circumstance_tool(query:str, callback:callable=None) -> str:
    query = query.strip(' "')
    return f'The temperature of {query} is 19 degrees. '

@custom_tool(
    name = "IOTControler",
    description = (
        "A tool for controling smart devices."
        "Useful for when you need to turn on or turn off a light, set up a air conditioning or other smart devices. "
        "Input should be the device name. "
    )
)
def iot_control_tool(query:str, callback:callable=None) -> str:
    query = query.strip(' "')
    return 'Operation success. '

@disk_cache(cache_path=os.path.join(CACHE_DIR, "closeai_generate_image.db"), expire_time=3600)
def generate_image(prompt):
    return api.closeai.generate_image(prompt)

@custom_tool(
    name = "OpenaiAIPainter",
    description = (
        "An AI painter. "
        "Useful for when you need to draw or paint or generate an image or a picture. With this tool, you are a painter and can paint or draw anything. "
        "Input should be a prompt for the AI painter depicting what image you want to draw. The output is the url of the image."
    )
)
def ai_paint(prompt:str, callback:callable=None) -> str:
    prompt = prompt.strip()
    if not prompt:
        return 'I got empty input and can do nothing.'

    urls = generate_image(prompt)
    if not urls:
        return "Error occured while drawing the image."

    res_url = urls[0]
    if callback:
        callback(prompt, res_url)
    return f"Here is the painted image's url:{res_url}"

@custom_tool(
    name="SendMessage",
    description="Message Sender. Useful for when you need to send a message to somebody. Input should be (user:the target user, mesage:the message to be sent). "
)
def send_message_tool(arg:str):
    to_user, message = arg.split(',', 1)
    print(f"sending message to \x1b[32m{to_user}\x1b[0m:{message}")
    return "The message has been sent successfully."

if __name__ == '__main__':
    print('custom tools:', get_all_custom_tool_names())

    def on_weather(query, result):
        print(f"on_weather()  query=[{query}]")
        print(f"on_weather() result=[{result}]")
        return result
    def on_ai_paint(query, result):
        print(f"on_ai_paint()  query=[{query}]")
        print(f"on_ai_paint() result=[{result}]")
        return result
    def on_search_text(query, result):
        print(f"on_search_text()  query=[{query}]")
        print(f"on_search_text() result=[{result}]")
        return result
    tools = load_tools(["AIPainter", "WeatherForecast", "BaiduSearchText"], {
        'AIPainter': on_ai_paint,
        'WeatherForecast': on_weather,
        'BaiduSearchText': on_search_text,
    })

    for tool in tools:
        print(tool.name)

